#include "calculator_functions.h"
#include "payment_method.h"
#include "utilities.h"
#include "tariff_log.h"

#include <sstream>
#include <QDateTime>
#include <qdebug.h>

double total_duration_min = 0.0f;
double total_cost = 0.0f;
bool overtime = false;

#ifdef _WIN32
inline struct tm* localtime_r(const time_t *clock, struct tm* result){
    if(!clock || !result) return NULL;
    memcpy(result,localtime(clock),sizeof(*result));
    return result;
}
#endif

/// <inheritdoc/>
std::string Calculator::GetDurationFromCost(Configuration* cfg,
                                            uint8_t payment_option,
                                            char const* start_datetime, // given in local time
                                            double price,
                                            bool nextDay,
                                            bool prepaid)
{
    // Get input date
    QDateTime inputDate = QDateTime::fromString(start_datetime,Qt::ISODate);

    // Get day of week
    int weekdayId = 0;
    weekdayId = Utilities::ZellersAlgorithm(inputDate.date().day(),inputDate.date().month(),inputDate.date().year());

    //Get min and max time defined in JSON
    double minMin = 0;
    minMin = cfg->PaymentOption.find(payment_option)->second.pop_min_time;

    double maxMin = 0;
    maxMin = cfg->PaymentOption.find(payment_option)->second.pop_max_time;

    double min_price = 0;
    min_price = cfg->PaymentOption.find(payment_option)->second.pop_min_price;
    if(price < min_price)
    {
        return "PARKING NOT ALLOWED";
    }

    if (minMin < 0) minMin = 0;
    if (maxMin < 0) maxMin = 0;
    if (minMin >= maxMin)
    {
        LOG_ERROR("Error: min_min cannot be greater or equal to max_min");
        return "PARKING NOT ALLOWED";
    }

    if (maxMin <= minMin)
    {
        LOG_ERROR("Error: max_min cannot be lower or equal than min_min");
        return "PARKING NOT ALLOWED";
    }


    // Get payment method
    uint8_t p_method = PaymentMethod::Undefined;
    p_method = payment_option;
    LOG_DEBUG("Payment method id: ", (unsigned)p_method);

    // Check special day
    double day_price = 0.0f;
    int current_special_day_id = -1;
    bool is_special_day = Utilities::CheckSpecialDay(cfg, inputDate.toString(Qt::ISODate).toStdString().c_str(), &current_special_day_id, &day_price);
    LOG_DEBUG("Special day: ", is_special_day);

    double money_left = price;
    double price_per_unit = 0.0f;

    QTime worktime_from;
    QTime worktime_to;

    if(is_special_day)
    {
        // Set special day price
        price_per_unit = Utilities::CalculatePricePerUnit(day_price);
        worktime_from = QTime::fromString(cfg->SpecialDaysWorktime.find(current_special_day_id)->second.pedwt_time_from.c_str());
        worktime_to = QTime::fromString(cfg->SpecialDaysWorktime.find(current_special_day_id)->second.pedwt_time_to.c_str());
    }
    else
    {
        // Set new price for the normal day
        day_price = cfg->PaymentRate.find(payment_option)->second.pra_price;
        price_per_unit = Utilities::CalculatePricePerUnit(day_price);

        // If no working day found, skip it (recursively call method again)
        size_t found = 0;
        found = cfg->WeekDaysWorktime.count(weekdayId);

        // When no workday found, go to next available day
        if(found <=0)
        {
            LOG_DEBUG("- No workday found, trying to find next available day");
            inputDate = inputDate.addDays(1);
            return GetDurationFromCost(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), money_left,true,prepaid);
        }
        worktime_from = QTime::fromString(cfg->WeekDaysWorktime.find(weekdayId)->second.pwd_time_from.c_str());
        worktime_to = QTime::fromString(cfg->WeekDaysWorktime.find(weekdayId)->second.pwd_time_to.c_str());
    }

    if (price_per_unit < 0) price_per_unit = 1.0f;
    LOG_DEBUG("Calculated price per minute: ", price_per_unit);

    if (price_per_unit < 0)
    {
        inputDate = inputDate.addDays(1);
        inputDate.setTime(worktime_from);
        return GetDurationFromCost(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), money_left, true);
    }

    // If overtime flag is set
    if (overtime || nextDay)
    {
        inputDate.setTime(worktime_from);
        overtime = false;
    }

    // Check prepaid
    if (!prepaid)
    {
        if ((inputDate.time() < worktime_from) || (inputDate.time() > worktime_to))
        {
            LOG_DEBUG("[STOP] * Ticket is not valid * ");
            return "PARKING NOT ALLOWED";
        }
    }
    else
    {
        LOG_DEBUG("* PREPAID MODE ACTIVE *");
        if (inputDate.time() < worktime_from)
        {
            inputDate.setTime(worktime_from);
        }
        else if(inputDate.time() > worktime_to)
        {
            LOG_DEBUG(" *** PREPAID *** Current time is past the time range end, searching for next available day");
            inputDate = inputDate.addDays(1);
            return GetDurationFromCost(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), money_left, true);
        }
    }

    while(true)
    {
        if(money_left <= 0) break;

        // Check year period
        bool isYearPeriodActive = false;

        //// Parse input date
        int dayCurrent = inputDate.date().day();
        int monthCurrent = inputDate.date().month();

        // Current date time
        int cdt = (monthCurrent * 100) + dayCurrent;

        multimap<int, ATBPeriodYear>::iterator year_period_itr;
        for (year_period_itr = cfg->YearPeriod.begin(); year_period_itr != cfg->YearPeriod.end(); ++year_period_itr)
        {
            int dStart = year_period_itr->second.pye_start_day;
            int dEnd = year_period_itr->second.pye_end_day;

            int mStart = year_period_itr->second.pye_start_month;
            int mEnd = year_period_itr->second.pye_end_month;

            int start = (mStart * 100) + dStart;
            int end = (mEnd * 100) + dEnd;

            if (cdt >= start && cdt <= end) {
                isYearPeriodActive = true;
                break;
            }
        }
        if (!isYearPeriodActive)
        {
            LOG_DEBUG("Year period is not valid");
            return "PARKING NOT ALLOWED";
        }

        if(total_duration_min > maxMin)
        {
            total_duration_min = maxMin;
            break;
        }

        // If reached end of worktime go to next day
        if(inputDate.time() >= worktime_to)
        {
            int carry_over_status = 0;
            carry_over_status = cfg->PaymentOption.find(payment_option)->second.pop_carry_over;
            if (carry_over_status < 1) break;

            inputDate = inputDate.addDays(1);
            overtime = true;
            return GetDurationFromCost(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), money_left ,true, prepaid);
        }

        if(money_left > 1)
        inputDate = inputDate.addSecs(60);

        if(price_per_unit > 0) total_duration_min +=1;
        money_left -= price_per_unit;

        //qDebug() <<"Timestamp:" << inputDate << ", total duration min: " << total_duration_min << ", money left = " << money_left;
    }

    if ((total_duration_min < minMin) || (price / price_per_unit) < minMin)
    {
        LOG_DEBUG("Total duration is lower than min_min");
        inputDate.time() = worktime_from;
        total_duration_min = 0;
    }

    double ret_val = 0;
    double calc_price = (int)total_duration_min - (int)price / price_per_unit;

    if (calc_price > 0 && total_duration_min > 0)
    {
        inputDate.addSecs(-(int)ceil(calc_price) * 60);
    }
    ret_val = total_duration_min;

    if(price >= min_price && total_duration_min >= minMin)
        qDebug() << "Valid until: " << inputDate.toString(Qt::ISODate);
    else  qDebug() << "Parking not allowed";

    if(ret_val < 0) ret_val = 0;
    qDebug() << "Duration: " << ret_val;
    if (ret_val <= 0) return "PARKING NOT ALLOWED";

    total_duration_min = 0;
    return inputDate.toString(Qt::ISODate).toStdString();
}

///////////////////////////////////////

/// <inheritdoc/>
double Calculator::GetCostFromDuration(Configuration* cfg, uint8_t payment_option, const char* start_datetime, double durationMin, bool nextDay, bool prepaid)
{
    // Get input date
    QDateTime inputDate = QDateTime::fromString(start_datetime,Qt::ISODate);

    // Get day of week
    int weekdayId = 0;
    weekdayId = Utilities::ZellersAlgorithm(inputDate.date().day(),inputDate.date().month(),inputDate.date().year());

    //Get min and max time defined in JSON
    double minMin = 0;
    minMin = cfg->PaymentOption.find(payment_option)->second.pop_min_time;

    double maxMin = 0;
    maxMin = cfg->PaymentOption.find(payment_option)->second.pop_max_time;

    if (minMin < 0) minMin = 0;
    if (maxMin < 0) maxMin = 0;
    if (minMin >= maxMin)
    {
        LOG_ERROR("Error: min_min cannot be greater or equal to max_min");
        return 0.0f;
    }
    if (maxMin <= minMin)
    {
        LOG_ERROR("Error: max_min cannot be lower or equal than min_min");
        return 0.0f;
    }

    // Check overtime
    if (!overtime)
    {
        if (durationMin > maxMin)
        {
            LOG_WARNING("Total duration is greater or equal to max_min");
            return maxMin;
        }

        if (durationMin < minMin)
        {
            LOG_WARNING("Total duration is lower or equal to min_min");
            return 0.0f;
        }
    }

    // Get payment method
    uint8_t p_method = PaymentMethod::Undefined;
    p_method = payment_option;
    LOG_DEBUG("Payment method id: ", (unsigned)p_method);

    // Check special day
    double day_price = 0.0f;
    int current_special_day_id = -1;
    bool is_special_day = Utilities::CheckSpecialDay(cfg, inputDate.toString(Qt::ISODate).toStdString().c_str(), &current_special_day_id, &day_price);
    LOG_DEBUG("Special day: ", is_special_day);

    total_duration_min = durationMin;
    LOG_DEBUG("Total min:", total_duration_min);

    double price_per_unit = 0.0f;
    QTime worktime_from;
    QTime worktime_to;

    if(is_special_day)
    {
        // Set special day price
        price_per_unit = Utilities::CalculatePricePerUnit(day_price);
        worktime_from = QTime::fromString(cfg->SpecialDaysWorktime.find(current_special_day_id)->second.pedwt_time_from.c_str());
        worktime_to = QTime::fromString(cfg->SpecialDaysWorktime.find(current_special_day_id)->second.pedwt_time_to.c_str());
    }
    else
    {
        // Set new price for the normal day
        day_price = cfg->PaymentRate.find(payment_option)->second.pra_price;
        price_per_unit = Utilities::CalculatePricePerUnit(day_price);

        // If no working day found, skip it (recursively call method again)
        size_t found = 0;
        found = cfg->WeekDaysWorktime.count(weekdayId);

        // When no workday found, go to next available day
        if(found <=0)
        {
            LOG_DEBUG("- No workday found, trying to find next available day");
            inputDate = inputDate.addDays(1);
            return floor(GetCostFromDuration(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), durationMin, true, prepaid));
        }
        worktime_from = QTime::fromString(cfg->WeekDaysWorktime.find(weekdayId)->second.pwd_time_from.c_str());
        worktime_to = QTime::fromString(cfg->WeekDaysWorktime.find(weekdayId)->second.pwd_time_to.c_str());
    }

    if (price_per_unit < 0) price_per_unit = 1.0f;
    LOG_DEBUG("Calculated price per minute: ", price_per_unit);

    if (price_per_unit == 0)
    {
        inputDate = inputDate.addDays(1);
        inputDate.setTime(worktime_from);
        return GetCostFromDuration(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), durationMin, true, prepaid);
    }

    // If overtime flag is set
    if (overtime || nextDay)
    {
        inputDate.setTime(worktime_from);
        overtime = false;
    }

    // Check prepaid
    if (!prepaid)
    {
        if ((inputDate.time() < worktime_from) || (inputDate.time() > worktime_to))
        {
            LOG_DEBUG("[STOP] * Ticket is not valid * ");
            return 0.0f;
        }
    }
    else
    {
        LOG_DEBUG("* PREPAID MODE ACTIVE *");
        if (inputDate.time() < worktime_from)
        {
            inputDate.setTime(worktime_from);
        }
        else if(inputDate.time() > worktime_to)
        {
            LOG_DEBUG(" *** PREPAID *** Current time is past the time range end, searching for next available day");
            inputDate = inputDate.addDays(1);
            return GetCostFromDuration(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), durationMin, true, prepaid);
        }
    }

    while(true)
    {
        if(total_duration_min <= 0) break;

        // Check year period
        bool isYearPeriodActive = false;

        //// Parse input date
        int dayCurrent = inputDate.date().day();
        int monthCurrent = inputDate.date().month();

        // Current date time
        int cdt = (monthCurrent * 100) + dayCurrent;

        multimap<int, ATBPeriodYear>::iterator year_period_itr;
        for (year_period_itr = cfg->YearPeriod.begin(); year_period_itr != cfg->YearPeriod.end(); ++year_period_itr)
        {
            int dStart = year_period_itr->second.pye_start_day;
            int dEnd = year_period_itr->second.pye_end_day;

            int mStart = year_period_itr->second.pye_start_month;
            int mEnd = year_period_itr->second.pye_end_month;

            int start = (mStart * 100) + dStart;
            int end = (mEnd * 100) + dEnd;

            if (cdt >= start && cdt <= end) {
                isYearPeriodActive = true;
                break;
            }
        }
        if (!isYearPeriodActive)
        {
            LOG_DEBUG("Year period is not valid");
            return 0.0f;
        }

        // Go to next day if minutes not spent
        if(inputDate.time() >= worktime_to)
        {
            int carry_over_status = 0;
            carry_over_status = cfg->PaymentOption.find(payment_option)->second.pop_carry_over;
            if (carry_over_status < 1) break;

            LOG_DEBUG("Reached end of worktime, searching for the next working day");
            inputDate = inputDate.addDays(1);
            overtime = true;
            return GetCostFromDuration(cfg, payment_option, inputDate.toString(Qt::ISODate).toStdString().c_str(), total_duration_min);
        }

        // Increment input date minutes for each monetary unit
        inputDate = inputDate.addSecs(60);
        total_duration_min -=1;
        total_cost += price_per_unit;

    }
    qDebug() << "Valid until:" << inputDate.toString(Qt::ISODate).toStdString().c_str();
    double ret_val = total_cost;
    total_cost = 0.0f;
    return ceil(ret_val);
}
